"""``metallurgy.constants``
=============================

Module providing some physical constants used in calculation functions.

"""


plankConstant = 6.63e-34
avogadroNumber = 6.022e23
idealGasConstant = 8.31
boltzmann = 1.38064852e-23
